from nonebot.log import logger
from nonebot.matcher import Matcher
from nonebot.plugin import PluginMetadata, inherit_supported_adapters, require

require("nonebot_plugin_alconna")
require("nonebot_plugin_htmlrender")

from nonebot_plugin_alconna import Alconna, Args, UniMessage, on_alconna

from .config import Config, plugin_config
from .render_pic import render
from .weather_data import CityNotFoundError, ConfigError, Weather

__plugin_meta__ = PluginMetadata(
    name="和风天气",
    description="和风天气图片显示插件",
    usage="天气地名 / 地名天气",
    type="application",
    homepage="https://github.com/kexue-z/nonebot-plugin-heweather",
    config=Config,
    supported_adapters=inherit_supported_adapters("nonebot_plugin_alconna"),
)


if plugin_config.debug:
    logger.debug("将会保存图片到 weather.png")


weather = on_alconna(Alconna("天气", Args["city", str]), block=True, priority=1)
weather.shortcut(r"^(?P<city>.+)天气$", {"args": ["{city}"], "fuzzy": False})
weather.shortcut(r"^天气(?P<city>.+)$", {"args": ["{city}"], "fuzzy": False})


@weather.handle()
async def _(matcher: Matcher, city: str):
    if not (plugin_config.qweather_apikey or plugin_config.qweather_use_jwt):
        raise ConfigError("请设置 QWEATHER_APIKEY 或 JWT ")

    if plugin_config.qweather_apitype is None:
        raise ConfigError("请设置 QWEATHER_APITYPE")

    w_data = Weather(city_name=city, api_type=plugin_config.qweather_apitype)
    try:
        await w_data.load_data()
    except CityNotFoundError:
        logger.warning(f"找不到城市: {city}")
        matcher.block = False
        await matcher.finish()

    img = await render(w_data)

    if plugin_config.debug:
        debug_save_img(img)

    await UniMessage.image(raw=img).send()


def debug_save_img(img: bytes) -> None:
    from io import BytesIO

    from PIL import Image

    logger.debug("保存图片到 weather.png")
    a = Image.open(BytesIO(img))
    a.save("weather.png", format="PNG")



from os import getenv

import nonebot
from nonebot.adapters.onebot.v11 import Adapter
import pytest


@pytest.fixture(scope="session", autouse=True)
def _():
    # 加载适配器
    driver = nonebot.get_driver()
    driver.register_adapter(Adapter)

    config = nonebot.get_driver().config

    config.qweather_apikey = getenv("QWEATHER_APIKEY")
    config.qweather_apitype = getenv("QWEATHER_APITYPE")
    config.qweather_forecase_days = getenv("QWEATHER_FORECASE_DAYS")

    # 加载插件
    nonebot.load_plugin("nonebot_plugin_heweather")



from datetime import datetime
from pathlib import Path
import platform

from nonebot_plugin_htmlrender import template_to_pic

from .config import plugin_config
from .model import Air, Daily, Hourly, HourlyType
from .weather_data import Weather


async def render(weather: Weather) -> bytes:
    template_path = str(Path(__file__).parent / "templates")

    air = None
    if weather.air:
        if weather.air.now:
            air = add_tag_color(weather.air.now)

    return await template_to_pic(
        template_path=template_path,
        template_name="weather.html",
        templates={
            "now": weather.now.now,
            "days": add_date(weather.daily.daily),
            "city": weather.city_name,
            "warning": weather.warning,
            "air": air,
            "hours": add_hour_data(weather.hourly.hourly),
        },
        pages={
            "viewport": {"width": 1000, "height": 300},
            "base_url": f"file://{template_path}",
        },
    )


def add_hour_data(hourly: list[Hourly]):
    min_temp = min([int(hour.temp) for hour in hourly])
    high = max([int(hour.temp) for hour in hourly])
    low = int(min_temp - (high - min_temp))
    for hour in hourly:
        date_time = datetime.fromisoformat(hour.fxTime)
        if platform.system() == "Windows":
            hour.hour = date_time.strftime("%#I%p")
        else:
            hour.hour = date_time.strftime("%-I%p")
        if high == low:
            hour.temp_percent = "100px"
        else:
            hour.temp_percent = f"{int((int(hour.temp) - low) / (high - low) * 100)}px"
    if plugin_config.qweather_hourlytype == HourlyType.current_12h:
        hourly = hourly[:12]
    if plugin_config.qweather_hourlytype == HourlyType.current_24h:
        hourly = hourly[::2]
    return hourly


def add_date(daily: list[Daily]):
    week_map = [
        "周日",
        "周一",
        "周二",
        "周三",
        "周四",
        "周五",
        "周六",
    ]

    for day in daily:
        date = day.fxDate.split("-")
        _year = int(date[0])
        _month = int(date[1])
        _day = int(date[2])
        week = int(datetime(_year, _month, _day, 0, 0).strftime("%w"))
        day.week = week_map[week] if day != 0 else "今日"
        day.date = f"{_month}月{_day}日"

    return daily


def add_tag_color(air: Air):
    color = {
        "优": "#95B359",
        "良": "#A9A538",
        "轻度污染": "#E0991D",
        "中度污染": "#D96161",
        "重度污染": "#A257D0",
        "严重污染": "#D94371",
    }
    air.tag_color = color[air.category]
    return air



import asyncio

from httpx import URL, AsyncClient, Response
from nonebot.log import logger

from .config import plugin_config
from .model import AirApi, DailyApi, HourlyApi, NowApi, WarningApi
from .types import APIError, CityNotFoundError, ConfigError
from .utils import get_jwt_token


class Weather:
    def __url__(self):
        self.host = URL(plugin_config.qweather_apihost)
        # self.url_geoapi = "https://geoapi.qweather.com/v2/city/"
        # if self.api_type == 2 or self.api_type == 1:
        #     self.url_weather_api = "https://api.qweather.com/v7/weather/"
        #     self.url_weather_warning = "https://api.qweather.com/v7/warning/now"
        #     self.url_air = "https://api.qweather.com/v7/air/now"
        #     self.url_hourly = "https://api.qweather.com/v7/weather/24h"

        #     logger.info("使用标准订阅API")

        # elif self.api_type == 0:
        #     self.url_weather_api = "https://devapi.qweather.com/v7/weather/"
        #     self.url_weather_warning = "https://devapi.qweather.com/v7/warning/now"
        #     self.url_air = "https://devapi.qweather.com/v7/air/now"
        #     self.url_hourly = "https://devapi.qweather.com/v7/weather/24h"

        #     logger.info("使用免费订阅API")
        # else:
        #     raise ConfigError(
        #         "api_type 必须是为 (int)0 -> 免费订阅, "
        #         "(int)1 -> 标准订阅, (int)2 -> 商业版"
        #         f"\n当前为: ({type(self.api_type)}){self.api_type}"
        #     )

    def _forecast_days(self):
        self.forecast_days = plugin_config.qweather_forecase_days
        if self.forecast_days:
            if self.api_type == 0 and not (3 <= self.forecast_days <= 7):
                raise ConfigError("api_type = 0 免费订阅 预报天数必须 3<= x <=7")

    def __init__(self, city_name: str, api_type: int = 0):
        self.city_name = city_name
        # self.apikey = api_key
        self.api_type = api_type
        self.__url__()

        self._forecast_days()

        # self.now: Optional[Dict[str, str]] = None
        # self.daily = None
        # self.air = None
        # self.warning = None
        self.__reference = "\n请参考: https://dev.qweather.com/docs/start/status-code/"

    async def load_data(self):
        self.city_id = await self._get_city_id()
        (
            self.now,
            self.daily,
            self.air,
            self.warning,
            self.hourly,
        ) = await asyncio.gather(
            self._now, self._daily, self._air, self._warning, self._hourly
        )
        self._data_validate()

    async def _get_data(self, url: URL, params: dict) -> Response:
        headers = {}

        if plugin_config.qweather_apikey:
            headers = {"X-QW-Api-Key": plugin_config.qweather_apikey}

        if plugin_config.qweather_use_jwt and not plugin_config.qweather_apikey:
            headers = {
                "Authorization": f"Bearer {get_jwt_token()}",
            }

        if not headers:
            raise ConfigError("请确保已经配置 apikey 或 jwt")

        async with AsyncClient() as client:
            res = await client.get(url, params=params, headers=headers)
        return res

    async def _get_city_id(self):
        url = self.host.join("/geo/v2/city/lookup")
        res = await self._get_data(
            url=url,
            params={"location": self.city_name, "number": 1},
        )

        res = res.json()

        if res["code"] == "404":
            raise CityNotFoundError()
        elif res["code"] != "200":
            raise APIError("错误! 错误代码: {}".format(res["code"]) + self.__reference)
        else:
            self.city_name = res["location"][0]["name"]
            return res["location"][0]["id"]

    def _data_validate(self):
        if self.now.code == "200" and self.daily.code == "200":
            pass
        else:
            raise APIError(
                "错误! 请检查配置! "
                f"错误代码: now: {self.now.code}  "
                f"daily: {self.daily.code}  "
                + "air: {}  ".format(self.air.code if self.air else "None")
                + "warning: {}".format(self.warning.code if self.warning else "None")
                + self.__reference
            )

    def _check_response(self, response: Response) -> bool:
        if response.status_code == 200:
            logger.debug(f"{response.json()}")
            return True
        else:
            raise APIError(f"Response code:{response.status_code}")

    @property
    async def _now(self) -> NowApi:
        url = self.host.join("/v7/weather/now")
        res = await self._get_data(
            url=url,
            params={"location": self.city_id},
        )
        self._check_response(res)
        return NowApi(**res.json())

    @property
    async def _daily(self) -> DailyApi:
        url = self.host.join(f"/v7/weather/{self.forecast_days}d")
        res = await self._get_data(
            url=url,
            params={"location": self.city_id},
        )
        self._check_response(res)
        return DailyApi(**res.json())

    @property
    async def _air(self) -> AirApi:
        url = self.host.join("/v7/air/now")
        res = await self._get_data(
            url=url,
            params={"location": self.city_id},
        )
        self._check_response(res)
        return AirApi(**res.json())

    @property
    async def _warning(self) -> WarningApi | None:
        url = self.host.join("/v7/warning/now")
        res = await self._get_data(
            url=url,
            params={"location": self.city_id},
        )
        self._check_response(res)
        return None if res.json().get("code") == "204" else WarningApi(**res.json())

    @property
    async def _hourly(self) -> HourlyApi:
        url = self.host.join("/v7/weather/24h")
        res = await self._get_data(
            url=url,
            params={"location": self.city_id},
        )
        self._check_response(res)
        return HourlyApi(**res.json())



from enum import IntEnum

from nonebot.compat import PYDANTIC_V2, ConfigDict
from pydantic import BaseModel


class Now(BaseModel):
    if PYDANTIC_V2:
        model_config = ConfigDict(extra="allow")
    else:

        class Config:
            extra = "allow"

    obsTime: str
    temp: str
    icon: str
    text: str
    windScale: str
    windDir: str
    humidity: str
    precip: str
    vis: str


class NowApi(BaseModel):
    if PYDANTIC_V2:
        model_config = ConfigDict(extra="allow")
    else:

        class Config:
            extra = "allow"

    code: str
    now: Now


class Daily(BaseModel):
    if PYDANTIC_V2:
        model_config = ConfigDict(extra="allow")
    else:

        class Config:
            extra = "allow"

    fxDate: str
    week: str | None = None
    date: str | None = None
    tempMax: str
    tempMin: str
    textDay: str
    textNight: str
    iconDay: str
    iconNight: str


class DailyApi(BaseModel):
    if PYDANTIC_V2:
        model_config = ConfigDict(extra="allow")
    else:

        class Config:
            extra = "allow"

    code: str
    daily: list[Daily]


class Air(BaseModel):
    if PYDANTIC_V2:
        model_config = ConfigDict(extra="allow")
    else:

        class Config:
            extra = "allow"

    category: str
    aqi: str
    pm2p5: str
    pm10: str
    o3: str
    co: str
    no2: str
    so2: str
    tag_color: str | None = None


class AirApi(BaseModel):
    if PYDANTIC_V2:
        model_config = ConfigDict(extra="allow")
    else:

        class Config:
            extra = "allow"

    code: str
    now: Air | None = None


class Warning(BaseModel):
    if PYDANTIC_V2:
        model_config = ConfigDict(extra="allow")
    else:

        class Config:
            extra = "allow"

    title: str
    type: str
    pubTime: str
    text: str


class WarningApi(BaseModel):
    if PYDANTIC_V2:
        model_config = ConfigDict(extra="allow")
    else:

        class Config:
            extra = "allow"

    code: str
    warning: list[Warning] | None = None


class Hourly(BaseModel):
    if PYDANTIC_V2:
        model_config = ConfigDict(extra="allow")
    else:

        class Config:
            extra = "allow"

    fxTime: str
    hour: str | None = None
    temp: str
    icon: str
    text: str
    temp_percent: str | None = None


class HourlyApi(BaseModel):
    if PYDANTIC_V2:
        model_config = ConfigDict(extra="allow")
    else:

        class Config:
            extra = "allow"

    code: str
    hourly: list[Hourly]


class HourlyType(IntEnum):
    current_12h = 1
    current_24h = 2



from time import time

from jwt import encode

from .config import plugin_config
from .types import ConfigError


def get_jwt_token():
    if not (
        plugin_config.qweather_jwt_sub
        and plugin_config.qweather_jwt_kid
        and plugin_config.qweather_jwt_private_key
    ):
        raise ConfigError(
            "请检查是否遗漏了 QWEATHER_JWT_SUB "
            "QWEATHER_JWT_KID QWEATHER_JWT_PRIVATE_KEY "
            "中的其中一项"
        )

    payload = {
        "iat": int(time()) - 30,
        "exp": int(time()) + 900,
        "sub": plugin_config.qweather_jwt_sub,
    }
    headers = {
        "kid": plugin_config.qweather_jwt_kid,
        "alg": "EdDSA",
    }

    encoded_jwt = encode(
        payload,
        key=plugin_config.qweather_jwt_private_key,
        algorithm="EdDSA",
        headers=headers,
    )
    return encoded_jwt



from nonebot.plugin import get_plugin_config
from pydantic import BaseModel, Field

from .model import HourlyType


class Config(BaseModel):
    qweather_apihost: str = Field(default="https://api.qweather.com")
    qweather_apikey: str | None = Field(
        default=None, deprecated="建议使用更安全的JWT key"
    )
    qweather_apitype: int | None = Field(default=None)
    qweather_hourlytype: HourlyType | None = Field(default=HourlyType.current_12h)
    qweather_forecase_days: int | None = Field(default=3)
    qweather_use_jwt: bool | None = Field(
        default=True, description="是否使用 JWT，默认 True"
    )
    qweather_jwt_sub: str | None = Field(
        default=None, description="JWT sub，即控制台中的项目ID"
    )
    qweather_jwt_private_key: str | None = Field(
        default=None, deprecated="JWT 私钥文本，需要自行生成"
    )

    qweather_jwt_kid: str | None = Field(
        default=None, description="JWT Key ID，即控制台中上传公钥后即可获取"
    )
    debug: bool | None = Field(default=False)


plugin_config: Config = get_plugin_config(Config)



from nonebot import get_driver
from nonebug import App
import pytest


@pytest.mark.asyncio
async def test_heweather(app: App):
    from io import BytesIO

    from PIL import Image

    from nonebot_plugin_heweather.render_pic import render
    from nonebot_plugin_heweather.weather_data import CityNotFoundError, Weather

    async with app.test_server():
        config = get_driver().config
        api_key = config.qweather_apikey
        api_type = int(config.qweather_apitype)
        w_data = Weather(city_name="beijing", api_key=api_key, api_type=api_type)
        try:
            await w_data.load_data()
        except CityNotFoundError:
            pass

        img = await render(w_data)
        a = Image.open(BytesIO(img))
        a.save("docs/weather.png", format="PNG")
